package com.sksamuel.soa.rest

import com.fasterxml.jackson.module.scala.DefaultScalaModule
import javax.ws.rs.core.{Response, MediaType, GenericType, UriBuilder}
import java.net.URI
import org.codehaus.jackson.map.ObjectMapper
import org.glassfish.jersey.client.ClientConfig
import javax.ws.rs.client.{ClientFactory, Entity, Invocation, WebTarget}
import org.codehaus.jackson.jaxrs.JacksonJsonProvider
import com.sksamuel.scoot.rest.{RemoteResponse, Logging}

class RestClient extends Logging {

    private val mapper = new ObjectMapper
    mapper.registerModule(DefaultScalaModule)

    private val cc = new ClientConfig(new JacksonJsonProvider(mapper))
    private val client = ClientFactory.newClient(cc)

    def resource(url: String): RichTarget = resource(URI.create(url))
    def resource(uri: URI): RichTarget = new RichTarget(client.target(uri))
    def resource(urlPattern: String, params: Any*): RichTarget = {
        val uri = UriBuilder.fromPath(String.format(urlPattern, params)).build()
        resource(uri)
    }
}

class RichTarget(var target: WebTarget) {

    def param(name: String, value: String) = {
        target = target.queryParam(name, value)
        this
    }

    def post(entity: Any, mediaType: MediaType) = new
        RichInvocation(target.request().buildPost(Entity.entity(entity, mediaType)))
    def post(entity: Any, mediaType: String) = new
        RichInvocation(target.request().buildPost(Entity.entity(entity, mediaType)))

    def put(entity: Any, mediaType: MediaType) = new
        RichInvocation(target.request().buildPut(Entity.entity(entity, mediaType)))
    def put(entity: Any, mediaType: String) = new
        RichInvocation(target.request().buildPut(Entity.entity(entity, mediaType)))

    def get = new RichInvocation(target.request().buildGet())
    def delete = new RichInvocation(target.request().buildDelete())
}

class RichInvocation(invocation: Invocation) extends Logging {

    def status: Int = _safeInvoke(invocation)
    def invoke[T](implicit manifest: Manifest[T]): RemoteResponse[T] =
        _safeInvoke(invocation, Left(manifest.erasure.asInstanceOf[Class[T]]))
    def invoke[T](c: Class[T]): RemoteResponse[T] = _safeInvoke(invocation, Left(c))
    def invoke[T](gt: GenericType[T]): RemoteResponse[T] = _safeInvoke(invocation, Right(gt))

    private def _failure(response: Response): RemoteResponse[Nothing] = {
        val entity = response.readEntity(classOf[String])
        RemoteResponse.error(response.getStatus, entity)
    }
    private def _content[T](response: Response, either: Either[Class[T], GenericType[T]]): RemoteResponse[T] = {
        val entity = either match {
            case Left(x) => {
                logger.debug("status 200 marshalling into {}", x)
                response.readEntity(x)
            }
            case Right(x) => {
                logger.debug("status 200 marshalling into {}", x)
                response.readEntity(x)
            }
        }
        RemoteResponse.ok(entity)
    }
    private def _safeInvoke[T](invocation: Invocation, either: Either[Class[T], GenericType[T]]): RemoteResponse[T] = {
        val response = invocation.invoke()
        logger.debug("Response [status={} headers={}]", response.getStatusInfo, response.getStringHeaders)
        try {
            response.getStatus match {
                case 200 => _content(response, either)
                case 204 => RemoteResponse.nocontent
                case i: Int => _failure(response)
            }
        } catch {
            case e: Exception => {
                logger.warn("Error marshalling content", e)
                RemoteResponse.error(400, e.getMessage)
            }
        } finally {
            response.close()
        }
    }
    private def _safeInvoke[T](invocation: Invocation): Int = {
        val response = invocation.invoke()
        logger.debug("Response {}", response)
        try {
            response.getStatus
        } catch {
            case e: Exception => {
                logger.error("Error invoking method {}", e)
                -1
            }
        } finally {
            response.close()
        }
    }
}