package com.sksamuel.scoot.rest

import javax.ws.rs.core.MediaType
import java.io.InputStream
import javax.ws.rs.client.{Invocation, WebTarget, ClientFactory}
import org.codehaus.jackson.map.ObjectMapper
import com.fasterxml.jackson.module.scala.DefaultScalaModule
import org.glassfish.jersey.client.ClientConfig
import org.codehaus.jackson.jaxrs.JacksonJsonProvider

/** @author Stephen Samuel
  *
  *         Http Client
  *
  **/
object ScootClient {

    private val mapper = new ObjectMapper
    mapper.registerModule(DefaultScalaModule)

    private val cc = new ClientConfig(new JacksonJsonProvider(mapper))
    private[rest] val client = ClientFactory.newClient(cc)

    def resource(path: String): Request = Request(path,
        Map.empty,
        NoEncoding,
        NoCredentials,
        Map.empty,
        Map.empty,
        Get,
        Array(),
        null,
        Nil)
}

case class Request(path: String,
                   headers: Map[String, String],
                   encoding: Encoding,
                   credentials: Credentials,
                   queryParams: Map[String, String],
                   formParams: Map[String, String],
                   method: Method,
                   body: Array[Byte],
                   bodyType: String,
                   types: Seq[String]) {

    def encode(encoding: Encoding) = copy(encoding = encoding)
    def authenticate(cred: Credentials) = copy(credentials = credentials)
    def header(key: String, value: String) = copy(headers = headers + ((key, value)))
    def userAgent(ua: String) = copy(headers = headers + ("user-agent" -> ua))
    def referer(ua: String) = copy(headers = headers + ("referer" -> ua))
    def accept(types: String*) = copy(types = types)
    def cookie(name: String, value: String) = copy(headers = headers + (name -> value))
    def entity(body: Any, mediaType: MediaType) = this
    def queryParam(name: String, value: Any) = copy(queryParams = queryParams + (name -> value.toString))
    def formParam(name: String, value: Any) = copy(formParams = formParams + (name -> value.toString))
    def post(method: Method) = copy(method = method)

    def sendReceive: Response = {

        val target = ScootClient.client.target(path)
        val f: (WebTarget, (String, String)) => WebTarget = (t, p) => t.queryParam(p._1, p._2)
        val target2 = queryParams.foldLeft(target)(f)
        val builder = target2.request(types: _*)

        val g: (Invocation.Builder, (String, String)) => Invocation.Builder = (b, p) => b.header(p._1, p._2)
        val builder2 = headers.foldLeft(builder)(g)

        val r = builder2.build(method.getClass.getName)
        val resp = r.invoke()
        Response(resp.getStatus,
            resp.getStatusInfo.getReasonPhrase,
            Map.empty,
            resp.getEntity.asInstanceOf[InputStream],
            resp.getEntityTag.getValue)
    }
}

case class Response(status: Int,
                    reasonPhrase: String,
                    headers: Map[String, String],
                    resp: InputStream,
                    entityTag: String) {
    def decode(encoding: Encoding): Response = this
    def header(name: String): String = headers(name)
    def cookie(name: String): Cookie = null
    def length = header("Content-Length")
    def contentType = header("Content-Type")
    def unmarshal[T: Manifest] = null
}

class Cookie {
    /**
     * Cookies using the default version correspond to RFC 2109.
     */
    val DEFAULT_VERSION: Int = 1
    var name: String = null
    var value: String = null
    var version: Int = 0
    var path: String = null
    var domain: String = null
}

abstract class Credentials
case class BasicCredentials(name: String, password: String) extends Credentials
object NoCredentials extends Credentials


